package com.lacunasoftware.flutter.crosspki;

import android.content.Context;

import androidx.annotation.NonNull;

import java.io.IOException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import io.flutter.embedding.engine.plugins.FlutterPlugin;
import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.MethodChannel.MethodCallHandler;
import io.flutter.plugin.common.MethodChannel.Result;
import io.flutter.plugin.common.PluginRegistry.Registrar;

/** CrosspkiPlugin */
public class CrossPkiPlugin implements FlutterPlugin, MethodCallHandler {
	/// The MethodChannel that will the communication between Flutter and native Android
	///
	/// This local reference serves to register the plugin with the Flutter Engine and unregister it
	/// when the Flutter Engine is detached from the Activity
	private MethodChannel channel;
	private Context context;

	@Override
	public void onAttachedToEngine(@NonNull FlutterPluginBinding flutterPluginBinding) {
		channel = new MethodChannel(flutterPluginBinding.getFlutterEngine().getDartExecutor(), "plugin.lacunasoftware.com/crosspki");
		channel.setMethodCallHandler(this);
		setContext(flutterPluginBinding.getApplicationContext());
	}

	// This static function is optional and equivalent to onAttachedToEngine. It supports the old
	// pre-Flutter-1.12 Android projects. You are encouraged to continue supporting
	// plugin registration via this function while apps migrate to use the new Android APIs
	// post-flutter-1.12 via https://flutter.dev/go/android-project-migration.
	//
	// It is encouraged to share logic between onAttachedToEngine and registerWith to keep
	// them functionally equivalent. Only one of onAttachedToEngine or registerWith will be called
	// depending on the user's project. onAttachedToEngine or registerWith must both be defined
	// in the same class.
	public static void registerWith(Registrar registrar) {
		final MethodChannel channel = new MethodChannel(registrar.messenger(), "plugin.lacunasoftware.com/crosspki");
		CrossPkiPlugin plugin = new CrossPkiPlugin();
		plugin.setContext(registrar.context());
		channel.setMethodCallHandler(plugin);
	}

	@Override
	public void onMethodCall(@NonNull MethodCall call, @NonNull Result result) {
		try {

			if (call.method.equals("listCertificatesWithKey")) {
				result.success(listCertificatesWithKey(call));

			} else if (call.method.equals("importPkcs12")) {
				importPkcs12(call);
				result.success(null);

			} else if (call.method.equals("importCertificate")) {
				importCertificate(call);
				result.success(null);

			} else if (call.method.equals("removeCertificate")) {
				removeCertificate(call);
				result.success(null);

			} else if (call.method.equals("generateRsaKeyPair")) {
				result.success(generateRsaKeyPair(call));

			} else if (call.method.equals("getKeyPair")) {
				result.success(getKeyPair(call));

			} else if (call.method.equals("signData")) {
				result.success(signData(call));

			} else if (call.method.equals("signHash")) {
				result.success(signHash(call));

			} else if (call.method.equals("keySignData")) {
				result.success(keySignData(call));

			} else if (call.method.equals("keySignHash")) {
				result.success(keySignHash(call));

			} else {
				result.notImplemented();
			}

		} catch (Exception e) {
			ErrorCodes errorCode = ErrorCodes.INTERNAL;
			if (e instanceof CrossPkiException) {
				errorCode = ((CrossPkiException)e).getCode();
			}
			result.error(errorCode.getValue(), e.getMessage(), null);
		}
	}

	@Override
	public void onDetachedFromEngine(@NonNull FlutterPluginBinding binding) {
		channel.setMethodCallHandler(null);
	}

	private void setContext(Context context) {
		this.context = context;
	}

	private ArrayList<Map<String, Object>> listCertificatesWithKey(MethodCall call) throws Exception {
		ArrayList<X509Certificate> certificates = DroidCertificateStore.listCertificatesWithKey();
		return Util.toMap(certificates);
	}

	private void importPkcs12(MethodCall call) throws Exception {
		byte[] content = call.argument("pkcs12Content");
		String password = call.argument("password");
		DroidCertificateStore.importPkcs12(content, password);
	}

	private void importCertificate(MethodCall call) throws Exception {
		byte[] certificateContent = call.argument("certificateContent");
		String keyId = call.argument("keyId");
		DroidCertificateStore.importCertificate(certificateContent, keyId);
	}

	private void removeCertificate(MethodCall call) throws CertificateException, NoSuchAlgorithmException, KeyStoreException, IOException {
		String thumbprint = call.argument("certificateThumbprint");
		DroidCertificateStore.removeCertificate(Util.base64Decode(thumbprint));
	}

	private Map<String, Object> generateRsaKeyPair(MethodCall call) throws Exception {
		int keySize = call.argument("keySize");
		MobileKeyPair keyPair = DroidKeyStore.generateRsaKeyPair(keySize, context);
		return Util.toMap(keyPair);
	}

	private Map<String, Object> getKeyPair(MethodCall call) throws Exception {
		String keyId = call.argument("keyId");
		PublicKey publicKey = DroidKeyStore.getPublicKey(keyId);
		MobileKeyPair keyPair = new MobileKeyPair(keyId, publicKey);
		return Util.toMap(keyPair);
	}

	private Map<String, Object> signData(MethodCall call) throws Exception {
		String certificateThumbprint = call.argument("certificateThumbprint");
		String digestAlgorithm = call.argument("digestAlgorithm");
		byte[] data = call.argument("data");
		String keyId = DroidCertificateStore.findCertificate(Util.base64Decode(certificateThumbprint));
		return commonSignData(keyId, digestAlgorithm, data);
	}

	private Map<String, Object> signHash(MethodCall call) throws Exception {
		String certificateThumbprint = call.argument("certificateThumbprint");
		String digestAlgorithm = call.argument("digestAlgorithm");
		byte[] hash = call.argument("hash");
		String keyId = DroidCertificateStore.findCertificate(Util.base64Decode(certificateThumbprint));
		return commonSignHash(keyId, digestAlgorithm, hash);
	}

	private Map<String, Object> keySignData(MethodCall call) throws Exception {
		String keyId = call.argument("keyId");
		String digestAlgorithm = call.argument("digestAlgorithm");
		byte[] data = call.argument("data");
		return commonSignData(keyId, digestAlgorithm, data);
	}

	private Map<String, Object> keySignHash(MethodCall call) throws Exception {
		String keyId = call.argument("keyId");
		String digestAlgorithm = call.argument("digestAlgorithm");
		byte[] hash = call.argument("hash");
		return commonSignHash(keyId, digestAlgorithm, hash);
	}

	private Map<String, Object> commonSignData(String keyId, String digestAlgorithm, byte[] data) throws Exception {
		byte[] signature = DroidKeyStore.signData(keyId, digestAlgorithm, data);
		HashMap<String, Object> map = new HashMap<>();
		map.put("signature", signature);
		return map;
	}

	private Map<String, Object> commonSignHash(String keyId, String digestAlgorithm, byte[] hash) throws Exception {
		byte[] signature = DroidKeyStore.signHash(keyId, digestAlgorithm, hash);
		HashMap<String, Object> map = new HashMap<>();
		map.put("signature", signature);
		return map;
	}

}
