package com.lacunasoftware.flutter.crosspki;

import java.io.IOException;
import java.math.BigInteger;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.Signature;
import java.security.cert.CertificateException;
import java.security.spec.AlgorithmParameterSpec;


import android.annotation.SuppressLint;
import android.content.Context;
import android.os.Build;
import android.security.keystore.KeyGenParameterSpec;
import android.security.keystore.KeyProperties;

public class DroidKeyStore {

	public static final String pendingCertificateCN = "CN=pending-certificate-wpki";

	@SuppressLint({"NewApi"})
	public static MobileKeyPair generateRsaKeyPair(int keySize, Context context) throws Exception {
		String alias = Util.createUniqueId();
		String subjectDN = formatPendingCetificateDN(alias);
		KeyPairGenerator kpg = KeyPairGenerator.getInstance(KeyProperties.KEY_ALGORITHM_RSA, "AndroidKeyStore");
		AlgorithmParameterSpec algorithmSpec = null;

		if (android.os.Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
			// API 23+
			algorithmSpec = new KeyGenParameterSpec.Builder(alias,
					KeyProperties.PURPOSE_SIGN | KeyProperties.PURPOSE_VERIFY | KeyProperties.PURPOSE_ENCRYPT | KeyProperties.PURPOSE_DECRYPT)
					.setDigests(
							KeyProperties.DIGEST_MD5,
							KeyProperties.DIGEST_NONE,
							KeyProperties.DIGEST_SHA1,
							KeyProperties.DIGEST_SHA224,
							KeyProperties.DIGEST_SHA256,
							KeyProperties.DIGEST_SHA384,
							KeyProperties.DIGEST_SHA512
					)
					.setSignaturePaddings(KeyProperties.SIGNATURE_PADDING_RSA_PKCS1)
					.setKeySize(keySize)
					.setCertificateSubject(new javax.security.auth.x500.X500Principal(subjectDN))
					.setCertificateSerialNumber(BigInteger.ONE)
					.build();

		} else if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.JELLY_BEAN_MR2) {
			// API 18+
			java.util.Date nowDate = new java.util.Date();
			algorithmSpec = new android.security.KeyPairGeneratorSpec.Builder(context)
					.setAlias(alias)
					.setSubject(new javax.security.auth.x500.X500Principal(subjectDN))
					.setSerialNumber(BigInteger.ONE)
					.setStartDate(nowDate)
					.setEndDate(new java.util.Date(nowDate.getYear() + 10, 1, 1))
					.setKeySize(keySize)
					.setKeyType(KeyProperties.KEY_ALGORITHM_RSA)
					.build();
		}

		kpg.initialize(algorithmSpec);
		KeyPair keyPair = kpg.generateKeyPair();
		return new MobileKeyPair(alias, keyPair.getPublic());
	}

	public static byte[] signData(String keyId, String digestAlgorithm, byte[] data) throws Exception {
		DigestAlgorithm digest = DigestAlgorithm.parse(digestAlgorithm);
		KeyStore ks = getLoadedAndroidKeyStore();
		KeyStore.PrivateKeyEntry keyEntry = getKeyEntry(ks, keyId);
		Signature sig = Signature.getInstance(digest.getWithSignatureAlgorithm(keyEntry.getPrivateKey().getAlgorithm()));
		sig.initSign(keyEntry.getPrivateKey());
		sig.update(data);
		return sig.sign();
	}

	public static byte[] signHash(String keyId, String digestAlgorithm, byte[] hash) throws Exception {
		DigestAlgorithm digest = DigestAlgorithm.parse(digestAlgorithm);
		KeyStore ks = getLoadedAndroidKeyStore();
		KeyStore.PrivateKeyEntry keyEntry = getKeyEntry(ks, keyId);
		Signature sig = Signature.getInstance("NONEwith" + keyEntry.getPrivateKey().getAlgorithm());
		sig.initSign(keyEntry.getPrivateKey());
		sig.update(DigestInfo.encode(digest, hash));
		return sig.sign();
	}

	public static PublicKey getPublicKey(String keyId) throws Exception {
		KeyStore ks = getLoadedAndroidKeyStore();
		KeyStore.PrivateKeyEntry keyEntry = getKeyEntry(ks, keyId);
		return keyEntry.getCertificate().getPublicKey();
	}

	public static String formatPendingCetificateDN(String keyId) {
		return pendingCertificateCN + "-" + keyId;
	}

	public static KeyStore getLoadedAndroidKeyStore() throws KeyStoreException, CertificateException, NoSuchAlgorithmException, IOException {
		KeyStore ks = KeyStore.getInstance("AndroidKeyStore");
		ks.load(null);
		return ks;
	}

	public static KeyStore.PrivateKeyEntry getKeyEntry(KeyStore keystore, String alias) throws Exception {
		KeyStore.Entry entry = keystore.getEntry(alias, null);
		if (entry == null) {
			throw new CrossPkiException(ErrorCodes.KEY_NOT_FOUND, "Key not found: " + alias);
		}
		if (!(entry instanceof KeyStore.PrivateKeyEntry)) {
			throw new Exception("Android Key entry is not an instance of a PrivateKeyEntry: " + alias);
		}
		return (KeyStore.PrivateKeyEntry)entry;
	}

	public static boolean hasPrivateKey(KeyStore keystore, String alias) {
		try {
			return getKeyEntry(keystore, alias) != null;
		} catch (Exception e) { }
		return false;
	}
}