import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:asn1lib/asn1lib.dart';
import 'package:flutter/services.dart';

class CrossPki {
  static const MethodChannel _channel =
      const MethodChannel('plugin.lacunasoftware.com/crosspki');

  static Future<List<PKCertificate>> listCertificatesWithKey() async {
    List<dynamic> certificates = await _handledInvoke('listCertificatesWithKey');
    return certificates
      .cast<Map<dynamic, dynamic>>()
      .map((c) => PKCertificate._fromMap(c))
      .toList();
  }

  static Future<void> importPkcs12(Uint8List pkcs12Content, String password) async {
    await _handledInvoke('importPkcs12', <String, dynamic> {
      'pkcs12Content': pkcs12Content,
      'password': password,
    });
  }

  static Future<void> importCertificate(Uint8List certificateContent, String keyId) async {
    await _handledInvoke('importCertificate', <String, dynamic> {
      'certificateContent': certificateContent,
      'keyId': keyId,
    });
  }

  static Future<void> removeCertificate(String certificateThumbprint) async {
    await _handledInvoke('removeCertificate', <String, dynamic> {
      'certificateThumbprint': certificateThumbprint,
    });
  }

  static Future<MobileKeyPair> generateRsaKeyPair(int keySize) async {
    Map<dynamic, dynamic> keyPair = await _handledInvoke('generateRsaKeyPair', <String, dynamic> {
      'keySize': keySize,
    });
    return MobileKeyPair._fromMap(keyPair, PKAlgorithm.rsa);
  }

  static Future<Uint8List> signData(String certificateThumbprint, String digestAlgorithm, Uint8List data) async {
    return await _signData(certificateThumbprint, DigestAlgorithm.getFromString(digestAlgorithm), data);
  }

  static Future<Uint8List> _signData(String certificateThumbprint, DigestAlgorithm digestAlgorithm, Uint8List data) async {
    Map<dynamic, dynamic> signatureResponse = await _handledInvoke('signData', <String, dynamic> {
      'certificateThumbprint': certificateThumbprint,
      'digestAlgorithm': digestAlgorithm.oid,
      'data': data,
    });
    return signatureResponse["signature"];
  }

  static Future<Uint8List> signHash(String certificateThumbprint, String digestAlgorithm, Uint8List hash) async {
    return await _signHash(certificateThumbprint, DigestAlgorithm.getFromString(digestAlgorithm), hash);
  }

  static Future<Uint8List> _signHash(String certificateThumbprint, DigestAlgorithm digestAlgorithm, Uint8List hash) async {
    Map<dynamic, dynamic> signatureResponse = await _handledInvoke('signHash', <String, dynamic> {
      'certificateThumbprint': certificateThumbprint,
      'digestAlgorithm': digestAlgorithm.oid,
      'hash': hash,
    });
    return signatureResponse["signature"];
  }

  static Future<Uint8List> keySignData(String keyId, String digestAlgorithm, Uint8List hash) async {
    return await _keySignData(keyId, DigestAlgorithm.getFromString(digestAlgorithm), hash);
  }

  static Future<Uint8List> _keySignData(String keyId, DigestAlgorithm digestAlgorithm, Uint8List data) async {
    Map<dynamic, dynamic> signatureResponse = await _handledInvoke('keySignData', <String, dynamic> {
      'keyId': keyId,
      'digestAlgorithm': digestAlgorithm.oid,
      'data': data,
    });
    return signatureResponse["signature"];
  }

  static Future<Uint8List> keySignHash(String keyId, String digestAlgorithm, Uint8List hash) async {
    return await _keySignHash(keyId, DigestAlgorithm.getFromString(digestAlgorithm), hash);
  }

  static Future<Uint8List> _keySignHash(String keyId, DigestAlgorithm digestAlgorithm, Uint8List hash) async {
    Map<dynamic, dynamic> signatureResponse = await _handledInvoke('keySignHash', <String, dynamic> {
      'keyId': keyId,
      'digestAlgorithm': digestAlgorithm.oid,
      'hash': hash,
    });
    return signatureResponse["signature"];
  }

  static Future<Uint8List> generateCsr(MobileKeyPair keyPair) async {
    var signatureAlgorithm = keyPair.algorithm.getSignatureAlgorithm(DigestAlgorithm.sha256);
    var crInfo = _Asn1Util.createCertificationRequestInfo(keyPair.publicKeyEncoded, keyPair.algorithm);
    var signature = await CrossPki._keySignData(keyPair.keyId, signatureAlgorithm.digestAlgorithm, crInfo.encodedBytes);
    return _Asn1Util.createCsr(crInfo, signatureAlgorithm, signature);
  }

  static Future<T> _handledInvoke<T>(String method, [dynamic arguments]) async {
    try {
      return await _channel.invokeMethod(method, arguments);
    } on PlatformException catch (e) {
      throw CrossPkiException._(e.code, e.message);
    }
  }
}

//-----------------------------------------------------------------------------
// Exception

enum CrossPkiErrorCodes {
  unknown,
  internal,
  invalidLicense,
  wrongPassword,
  platformNotSupported,
  algorithmNotSupported,
  unknownAlgorithm,
  certificateNotFound,
  keyNotFound,
}

class CrossPkiException implements Exception {

  CrossPkiErrorCodes _code;
  CrossPkiErrorCodes get code => _code;

  final String message;

  CrossPkiException._(String codeStr, this.message) {
    _code = CrossPkiErrorCodes.values.firstWhere((c) => c.toString() == 'CrossPkiErrorCodes.$codeStr', orElse: () => CrossPkiErrorCodes.unknown);
  }

  CrossPkiException(CrossPkiErrorCodes code, this.message) {
    _code = code;
  }

  @override
  String toString() => '[$code]: $message';
}

//-----------------------------------------------------------------------------
// Models

class PKCertificate {

  final String thumbprint;
  final String subjectName;
  final String issuerName;
  final String subjectDisplayName;
  final String issuerDisplayName;
  final DateTime validityStart;
  final DateTime validityEnd;
  final Uint8List encoded;

  PKCertificate._({
    this.thumbprint,
    this.subjectName,
    this.issuerName,
    this.subjectDisplayName,
    this.issuerDisplayName,
    this.validityStart,
    this.validityEnd,
    this.encoded,
  });

  static PKCertificate _fromMap(Map map) {
    return PKCertificate._(
      thumbprint:         map['thumbprint'],
      subjectName:        map['subjectName'],
      issuerName:         map['issuerName'],
      subjectDisplayName: map['subjectDisplayName'],
      issuerDisplayName:  map['issuerDisplayName'],
      validityStart:      DateTime.parse(map['validityStart']),
      validityEnd:        DateTime.parse(map['validityEnd']),
      encoded:            map['encoded'],
    );
  }
}

// ----------------------------------------------------------------------------

class MobileKeyPair {

  final PKAlgorithm algorithm;
  final String keyId;
  final Uint8List publicKeyEncoded;

  MobileKeyPair._({
    this.keyId,
    this.publicKeyEncoded,
    this.algorithm,
  });

  static MobileKeyPair _fromMap(Map map, PKAlgorithm algorithm) {
    return MobileKeyPair._(
      keyId: map['keyId'],
      publicKeyEncoded: _Asn1Util.getPublicKeyEncoded(map['publicKeyEncoded']),
      algorithm: algorithm,
    );
  }
}

// ----------------------------------------------------------------------------
// Algorithms

abstract class Algorithm {
  final String name;
  final String oid;

  const Algorithm(this.name, this.oid);

  @override
  String toString() => '$name ($oid)';

  @override
  bool operator ==(other) => other is Algorithm && oid == other.oid;

  static String _canonString(String value) {
    return value.toLowerCase().trim().replaceAll('-', '').replaceAll(' ', '');
  }
}

// ----------------------------------------------------------------------------

class PKAlgorithm extends Algorithm {

  const PKAlgorithm._(String name, String oid): super(name, oid);

  static const rsa = const PKAlgorithm._('RSA', '1.2.840.113549.1.1.1');

  SignatureAlgorithm getSignatureAlgorithm(DigestAlgorithm digestAlgorithm) {
    return SignatureAlgorithm._allAlgorithms.firstWhere((a) => a.pkAlgorithm == this && a.digestAlgorithm == digestAlgorithm);
  }
}

// ----------------------------------------------------------------------------

class DigestAlgorithm extends Algorithm {

  const DigestAlgorithm._(String name, String oid): super(name, oid);

  static const md5 = const DigestAlgorithm._('MD5', '1.2.840.113549.2.5');
  static const sha1 = const DigestAlgorithm._('SHA-1', '1.3.14.3.2.26');
  static const sha224 = const DigestAlgorithm._('SHA-224', '2.16.840.1.101.3.4.2.4');
  static const sha256 = const DigestAlgorithm._('SHA-256', '2.16.840.1.101.3.4.2.1');
  static const sha384 = const DigestAlgorithm._('SHA-384', '2.16.840.1.101.3.4.2.2');
  static const sha512 = const DigestAlgorithm._('SHA-512', '2.16.840.1.101.3.4.2.3');

  static List<DigestAlgorithm> _allAlgorithms = [
    md5, sha1, sha224, sha256, sha384, sha512
  ];

  static DigestAlgorithm getFromString(String value) {
    final d = _allAlgorithms.firstWhere((element) => element.oid == value || Algorithm._canonString(value) == Algorithm._canonString(element.name), orElse: () => null);
    if (d == null) {
      throw CrossPkiException(CrossPkiErrorCodes.unknownAlgorithm, 'Could not parse digest algorithm: $value');
    }
    return d;
  }
}

// ----------------------------------------------------------------------------

class SignatureAlgorithm extends Algorithm {

  final PKAlgorithm pkAlgorithm;
  final DigestAlgorithm digestAlgorithm;

  const SignatureAlgorithm._(this.pkAlgorithm, this.digestAlgorithm, String name, String oid): super(name, oid);

  static const md5WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.md5, 'MD5withRSA', '1.2.840.113549.1.1.4');
  static const sha1WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.sha1, 'SHA1withRSA', '1.2.840.113549.1.1.5');
  static const sha224WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.sha224, 'SHA224withRSA', '1.2.840.113549.1.1.14');
  static const sha256WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.sha256, 'SHA256withRSA', '1.2.840.113549.1.1.11');
  static const sha384WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.sha384, 'SHA384withRSA', '1.2.840.113549.1.1.12');
  static const sha512WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.sha512, 'SHA512withRSA', '1.2.840.113549.1.1.13');

  static List<SignatureAlgorithm> _allAlgorithms = [
    md5WithRsa, sha1WithRsa, sha224WithRsa, sha256WithRsa, sha384WithRsa, sha512WithRsa
  ];
}

// ----------------------------------------------------------------------------

class _Asn1Util {

  static ASN1Sequence createCertificationRequestInfo(Uint8List publicKey, PKAlgorithm publicKeyAlgorithm) {
    var certificationRequestInfo = ASN1Sequence();
    // version
    certificationRequestInfo.add(ASN1Integer.fromInt(0));
    // subject
    certificationRequestInfo.add(ASN1Object.fromBytes(_subjectNameEncoded));
    // subjectPKInfo
    certificationRequestInfo.add(createSubjectPublicKeyInfo(publicKey, publicKeyAlgorithm));
    // attributes
    certificationRequestInfo.add(ASN1Null(tag: 0xA0));
    return certificationRequestInfo;
  }

  static Uint8List createCsr(ASN1Sequence certificationRequestInfo, SignatureAlgorithm signatureAlgorithm, Uint8List signature) {
    var csr = ASN1Sequence();
    // certificationRequestInfo;
    csr.add(certificationRequestInfo);
    // signatureAlgorithm
    csr.add(createAlgorithmIdentifier(signatureAlgorithm));
    // signature
    csr.add(ASN1BitString(signature.toList()));
    return csr.encodedBytes;
  }

  static ASN1Sequence createSubjectPublicKeyInfo(Uint8List publicKey, PKAlgorithm publicKeyAlgorithm) {
    var subjectPublicKeyInfo = ASN1Sequence();
    // algorithm
    subjectPublicKeyInfo.add(createAlgorithmIdentifier(publicKeyAlgorithm));
    // subjectPublicKey
    subjectPublicKeyInfo.add(ASN1BitString(publicKey.toList()));
    return subjectPublicKeyInfo;
  }

  static ASN1Sequence createAlgorithmIdentifier(Algorithm algorithm) {
    var algorithmSeq = ASN1Sequence();
    algorithmSeq.add(ASN1ObjectIdentifier.fromComponentString(algorithm.oid));

    if (algorithm == PKAlgorithm.rsa || (algorithm is SignatureAlgorithm && algorithm.pkAlgorithm == PKAlgorithm.rsa)) {
      algorithmSeq.add(ASN1Null(tag: 0x05));
      return algorithmSeq;
    }

    throw CrossPkiException(CrossPkiErrorCodes.algorithmNotSupported, 'Algorithm not supported: $algorithm');
  }

  static Uint8List getPublicKeyEncoded(Uint8List platformPublicKeyResult) {
    if (Platform.isIOS) {
      return platformPublicKeyResult;
    }

    if (Platform.isAndroid) {
      var subjectPublicKeyInfo = ASN1Sequence.fromBytes(platformPublicKeyResult);
      var bitString = subjectPublicKeyInfo.elements[1] as ASN1BitString;
      return bitString.contentBytes();
    }

    throw CrossPkiException(CrossPkiErrorCodes.platformNotSupported, 'Platform not supported: ${Platform.operatingSystem}');
  }

  static Uint8List _subjectNameEncoded = base64Decode("MB8xHTAbBgNVBAMTFENyb3NzUGtpLUZsdXR0ZXItS2V5");
}